package pps.core.base.service;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.text.CharSequenceUtil;
import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import pps.core.base.entity.BaseDataEnt;
import pps.core.base.entity.BaseDataImportLogEnt;
import pps.core.base.entity.BaseModelValEnt;
import pps.core.base.mapper.BaseDataImportLogMapper;
import pps.core.base.mapper.BaseDataMapper;
import pps.core.base.mapper.BaseModelValMapper;
import pps.core.base.service.data.base_data.*;
import pps.core.common.constant.BusinessConstant;
import xstartup.annotation.XService;
import xstartup.annotation.XText;
import xstartup.base.XContext;
import xstartup.base.data.CustomQueryInput;
import xstartup.base.util.XCopyUtils;
import xstartup.base.util.XHttpUtils;
import xstartup.base.util.XJsonUtils;
import xstartup.data.XListResult;
import xstartup.data.XPageResult;
import xstartup.data.XServiceResult;
import xstartup.data.XSingleResult;
import xstartup.error.XError;
import xstartup.feature.api.annotation.XApiGet;
import xstartup.feature.api.annotation.XApiPost;
import xstartup.feature.mybatis.helper.XMapperHelper;

import java.math.BigDecimal;
import java.util.*;

@XService
public class BaseDataService {
    @XApiPost
    public XServiceResult createBaseData(XContext context, CreateBaseDataInput input) {
        BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
        BaseDataEnt entity = new BaseDataEnt();
        XCopyUtils.copyObject(input, entity);
        mapper.insert(entity);
        return XServiceResult.OK;
    }

    @XApiPost
    public XServiceResult updateBaseData(XContext context, UpdateBaseDataInput input) {
        BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
        BaseDataEnt entity = mapper.selectOne(new LambdaQueryWrapper<BaseDataEnt>()
                .eq(BaseDataEnt::getId, input.getId()));
        if (ObjectUtil.isNull(entity)) {
            return XServiceResult.error(context, XError.NotFound);
        }
        XCopyUtils.copyObject(input, entity);
        mapper.updateById(entity);
        return XServiceResult.OK;
    }

    @XApiPost
    public XServiceResult deleteBaseData(XContext context, DeleteBaseDataInput input) {
        BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
        BaseDataEnt entity = mapper.selectOne(new LambdaQueryWrapper<BaseDataEnt>()
                .eq(BaseDataEnt::getId, input.getId()));
        if (ObjectUtil.isNull(entity)) {
            return XServiceResult.error(context, XError.NotFound);
        }
        mapper.deleteById(entity);
        return XServiceResult.OK;
    }

    @XApiGet
    public XSingleResult<GetBaseDataOutput> getBaseData(XContext context, GetBaseDataInput input) {
        BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
        BaseDataEnt entity = mapper.selectOne(new LambdaQueryWrapper<BaseDataEnt>()
                .eq(BaseDataEnt::getId, input.getId()));
        if (ObjectUtil.isNull(entity)) {
            return XSingleResult.error(context, XError.NotFound);
        }
        GetBaseDataOutput output = new GetBaseDataOutput();
        XCopyUtils.copyObject(entity, output);
        return XSingleResult.success(output);
    }

    @XApiGet
    public XListResult<QueryBaseDataOutput> queryBaseDataCopy(XContext context, GetBaseDataInput input) {
        BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
        LambdaQueryWrapper<BaseDataEnt> queryWrapper = new LambdaQueryWrapper<BaseDataEnt>()
                .eq(BaseDataEnt::getLineId, input.getLineId())
                .orderByAsc(BaseDataEnt::getDataDate);
        Date beginTime;
        Date endTime;
        if (CharSequenceUtil.isBlank(input.getDataDate())) {
            DateTime date = DateUtil.date();
            beginTime = DateUtil.beginOfDay(date);
            endTime = DateUtil.endOfDay(date);
            queryWrapper.ge(BaseDataEnt::getDataDate, beginTime)
                    .le(BaseDataEnt::getDataDate, endTime);
        } else {
            if (CharSequenceUtil.contains(input.getDataDate(), '@')) {
                String[] dataDates = CharSequenceUtil.splitToArray(input.getDataDate(), '@');
                beginTime = DateUtil.beginOfDay(DateUtil.parseDateTime(dataDates[0]));
                endTime = DateUtil.endOfDay(DateUtil.parseDateTime(dataDates[1]));
                queryWrapper.ge(BaseDataEnt::getDataDate, beginTime)
                        .le(BaseDataEnt::getDataDate, endTime);
            } else {
                beginTime = DateUtil.beginOfDay(DateUtil.parseDateTime(input.getDataDate()));
                endTime = DateUtil.endOfDay(DateUtil.parseDateTime(input.getDataDate()));
                queryWrapper.ge(BaseDataEnt::getDataDate, beginTime)
                        .le(BaseDataEnt::getDataDate, endTime);
            }
        }
        List<BaseDataEnt> list = mapper.selectList(queryWrapper);
        CalcBaseDataInput calcBaseDataInput = new CalcBaseDataInput("", input.getKeyName(), input.getLineId(), "train_pre", DateUtil.formatDateTime(beginTime), DateUtil.formatDateTime(endTime));
        Map<String, Object> map = getCalcData(context, calcBaseDataInput);
        list.forEach(item -> {
            String dataDate = DateUtil.formatDateTime(item.getDataDate());
            if (map.containsKey(dataDate)) {
                item.setPredictedPower(new BigDecimal(map.get(dataDate).toString()));
            }
        });
        return XListResult.success(XCopyUtils.copyNewList(list, QueryBaseDataOutput.class));
    }

    @XApiGet
    public XListResult<QueryBaseDataOutput> queryBaseData(XContext context, GetBaseDataInput input) {
        BaseDataImportLogMapper logMapper = context.getBean(BaseDataImportLogMapper.class);
        BaseDataImportLogEnt ent = logMapper.selectOne(new LambdaQueryWrapper<BaseDataImportLogEnt>()
                .eq(BaseDataImportLogEnt::getLineId, input.getLineId())
                .orderByDesc(BaseDataImportLogEnt::getCreateTime)
                .last(BusinessConstant.LAST_LIMIT));
        List<QueryBaseDataOutput> outputs = new ArrayList<>();
        if (ObjectUtil.isNotNull(ent)) {
            BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
            List<BaseDataEnt> list = mapper.selectList(new LambdaQueryWrapper<BaseDataEnt>()
                    .ge(BaseDataEnt::getDataDate, ent.getBeginDate())
                    .le(BaseDataEnt::getDataDate, ent.getEndDate())
                    .eq(BaseDataEnt::getLineId, input.getLineId())
                    .orderByAsc(BaseDataEnt::getDataDate));
            outputs = XCopyUtils.copyNewList(list, QueryBaseDataOutput.class);
        }
        return XListResult.success(outputs);
    }

    @XApiGet
    @XText("获取预测数据")
    public XListResult<GetBaseDataOutput> queryPreBaseData(XContext context, GetBaseDataInput input) {
        BaseDataImportLogMapper logMapper = context.getBean(BaseDataImportLogMapper.class);
        BaseDataImportLogEnt ent = logMapper.selectOne(new LambdaQueryWrapper<BaseDataImportLogEnt>()
                .eq(BaseDataImportLogEnt::getLineId, input.getLineId())
                .orderByDesc(BaseDataImportLogEnt::getCreateTime)
                .last(BusinessConstant.LAST_LIMIT));
        List<GetBaseDataOutput> outputs = new ArrayList<>();
        if (ObjectUtil.isNotNull(ent)) {
            BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
            List<BaseDataEnt> list = mapper.selectList(new LambdaQueryWrapper<BaseDataEnt>()
                    .ge(BaseDataEnt::getDataDate, ent.getBeginDate())
                    .le(BaseDataEnt::getDataDate, ent.getEndDate())
                    .eq(BaseDataEnt::getLineId, input.getLineId())
                    .orderByAsc(BaseDataEnt::getDataDate));
            Map<String, Object> map = getCalcData(context, new CalcBaseDataInput("", input.getKeyName(), input.getLineId(), "train_pre", DateUtil.formatDateTime(ent.getBeginDate()), DateUtil.formatDateTime(ent.getEndDate())));
            list.forEach(item -> {
                String dataDate = DateUtil.formatDateTime(item.getDataDate());
                if (map.containsKey(dataDate)) {
                    item.setPredictedPower(new BigDecimal(map.get(dataDate).toString()));
                }
            });
            outputs = XCopyUtils.copyNewList(list, GetBaseDataOutput.class);
        }
        return XListResult.success(outputs);
    }

    @XApiGet
    public XPageResult<QueryBaseDataOutput> queryBaseDataList(XContext context, QueryBaseDataInput input) {
        BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
        QueryWrapper<BaseDataEnt> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda()
                .eq(BaseDataEnt::getLineId, input.getLineId())
                .eq(BaseDataEnt::getDataDate, input.getDataDate());
        return XMapperHelper.query(mapper, input, queryWrapper, QueryBaseDataOutput.class);
    }

    @XApiGet
    public XPageResult<QueryBaseDataOutput> dynamicQueryBaseData(XContext context, CustomQueryInput input) {
        BaseDataMapper mapper = context.getBean(BaseDataMapper.class);
        return XMapperHelper.query(mapper, input, BaseDataEnt.class, QueryBaseDataOutput.class);
    }

    @XText("获取预算")
    @XApiGet
    public XServiceResult calcBaseData(XContext context, CalcBaseDataInput calcBaseDataInput) {
        getCalcData(context, calcBaseDataInput);
        return XServiceResult.OK;
    }

    public Map<String, Object> getCalcData(XContext context, CalcBaseDataInput calcBaseDataInput) {
        Map<String, Object> returnMap = new HashMap<>();
        try {
            String jsonString = XJsonUtils.toJson(calcBaseDataInput);
            context.getLogger().info(jsonString);
            String result = XHttpUtils.postAsJson("http://121.36.43.217:8098/model/predict", jsonString, 10000);
            returnMap = XJsonUtils.toMap(result);
            return returnMap;
        } catch (Exception e) {
            return returnMap;
        }
    }

    public void doTrainData(CalcBaseDataInput calcBaseDataInput) {
        try {
            String jsonString = XJsonUtils.toJson(calcBaseDataInput);
            XHttpUtils.postAsJson("http://121.36.43.217:8098/model/train", jsonString, 10000);
        } catch (Exception e) {
            throw new RuntimeException("调用预测接口失败");
        }
    }

    @XApiPost
    @XText("预测算法训练")
    public XServiceResult trainBaseData(XContext context, TrainBaseModelValInput input) {
        BaseDataImportLogMapper logMapper = context.getBean(BaseDataImportLogMapper.class);
        BaseDataImportLogEnt ent = logMapper.selectOne(new LambdaQueryWrapper<BaseDataImportLogEnt>()
                .eq(BaseDataImportLogEnt::getLineId, input.getLineId())
                .orderByDesc(BaseDataImportLogEnt::getCreateTime)
                .last(BusinessConstant.LAST_LIMIT));
        if (ObjectUtil.isNull(ent))
            return XServiceResult.OK;
        if (CollUtil.isNotEmpty(input.getKeyNames())) {
            for (String keyName : input.getKeyNames()) {
                CalcBaseDataInput calcBaseDataInput = new CalcBaseDataInput("", keyName, input.getLineId(), "train_val", DateUtil.formatDateTime(ent.getBeginDate()), DateUtil.formatDateTime(ent.getEndDate()));
                doTrainData(calcBaseDataInput);
            }
        } else {
            CalcBaseDataInput calcBaseDataInput = new CalcBaseDataInput("", input.getKeyName(), input.getLineId(), "train_val", DateUtil.formatDateTime(ent.getBeginDate()), DateUtil.formatDateTime(ent.getEndDate()));
            doTrainData(calcBaseDataInput);
        }
        return XServiceResult.OK;
    }

    @XApiGet
    @XText("获取最新算法训练数据")
    public XSingleResult<TrainBaseModelValOutput> getTrainBaseData(XContext context, GetTrainBaseModelValInput input) {
        BaseModelValMapper baseModelValMapper = context.getBean(BaseModelValMapper.class);
        if (ObjectUtil.isNull(input.getLineId())) {
            return XSingleResult.error(500, "请选择线路");
        }
        if (CharSequenceUtil.isBlank(input.getKeyNames())) {
            return XSingleResult.error(500, "请选择模型");
        }
        BaseModelValEnt baseModelValEnt = baseModelValMapper.selectOne(new LambdaQueryWrapper<BaseModelValEnt>()
                .eq(BaseModelValEnt::getLineId, input.getLineId())
                .eq(BaseModelValEnt::getKeyname, input.getKeyNames())
                .orderByDesc(BaseModelValEnt::getCreatetime).last(BusinessConstant.LAST_LIMIT));
        TrainBaseModelValOutput output = XCopyUtils.copyNewObject(baseModelValEnt, TrainBaseModelValOutput.class);
        if (ObjectUtil.isNotNull(baseModelValEnt) && CharSequenceUtil.isAllNotBlank(output.getYtrue(), output.getYpred())) {
            Map<String, Object> ytrueMap = XJsonUtils.toMap(output.getYtrue().replace("\\", ""));
            Map<String, Object> yPreMap = XJsonUtils.toMap(output.getYpred().replace("\\", ""));
            List<TrainBaseModelValDetailOutput> list = new ArrayList<>();
            String key;
            for (Map.Entry<String, Object> entry : ytrueMap.entrySet()) {
                key = entry.getKey();
                TrainBaseModelValDetailOutput detailOutput = new TrainBaseModelValDetailOutput();
                Date date = new Date(Long.parseLong(key));
                detailOutput.setKey(DateUtil.formatDateTime(date));
                detailOutput.setPower(entry.getValue().toString());
                if (yPreMap.containsKey(key)) {
                    detailOutput.setPredictedPower(yPreMap.get(key).toString());
                }
                list.add(detailOutput);
            }
            output.setList(list);
        } else {
            output = new TrainBaseModelValOutput();
        }
        return XSingleResult.success(output);
    }
}